import { liftF } from './fn.js'
import { Algebra, BaseSet, Monad, Semigroupoid } from './index.js'

/** @import { InferredMorphism, Morphism } from './types.js' */

/**
 * @template T
 * @extends BaseSet
 * @mixes Monad
 */
export class Suspend extends Algebra(Semigroupoid, Monad) {
  _value
  #fn

  /**
   * @param {T} value 
   * @param {InferredMorphism<T>} fn
   */
  constructor(value, fn) {
    super()
    this._value = value
    this.#fn = fn
  }

  /** @returns {this is Pure<T>} */
  isPure() { return false }

  /** @returns {this is Suspend<T>} */
  isSuspend() { return true }

  /**
   * @template U
   * @param {Morphism<T, Free<U>>} f 
   * @returns {Suspend<T>}
   */
  chain(f) {
    return new Suspend(this._value, x => this.#fn(x).chain(f))
  }

  /**
   * @template U
   * @param {Morphism<T, U>} g
   * @returns {Suspend<T>}
   */
  map(g) {
    return new Suspend(this._value, x => this.#fn(x).map(g))
  }

  /**
   * @template U
   * @param {Morphism<T, U>} g
   * @returns {Suspend<T>}
   */
  then(g) {
    return this.map(g)
  }

  /**
   * @template T
   * @param {InferredMorphism<T>} other 
   */
  compose(other) {
    return this.chain(() => other)
  }

  step() {
    return this.#fn(this._value)
  }

  run() {
    return this.step().run()
  }
}

/** @template T */
export class Pure extends Algebra(Monad) {
  _value

  /** @param {T} value */
  constructor(value) {
    super()
    this._value = value
  }

  /** @returns {this is Pure<T>} */
  isPure() { return true }

  /** @returns {this is Suspend<T>} */
  isSuspend() { return false }

  /**
   * @template U
   * @param {Morphism<T, Pure<U>>} f 
   * @returns {Pure<U>}
   */
  chain(f) {
    return f(this._value)
  }

  /**
   * @template U
   * @param {Morphism<T, U>} f 
   * @returns {Pure<U>}
   */
  map(f) {
    return this.chain(x => pure(f(x)))
  }

  /**
   * @template R
   * @param {Morphism<T, R>} f 
   * @returns {Pure<R>}
   */
  then(f) {
    return this.map(f)
  }

  /**
   * @template T
   * @param {InferredMorphism<T>} other 
   */
  compose(other) {
    return this.chain(() => pure(other))
  }

  step() {
    return this._value
  }

  run() {
    return this._value
  }
}

/**
 * @template T
 * @type {Pure<T> | Suspend<T>} Free
 */
export class Free {
  /**
   * @template T
   * @param {T} value
   * @returns {Free<T>}
   */
  static of(value) {
    return liftF(value)
  }
}

/**
 * @template T
 * @param {T} value
 * @returns {Pure<T>}
 */
export const pure = value => new Pure(value)

/**
 * @template T
 * @param {T} value
 * @param {InferredMorphism<T>} f
 * @returns {Suspend<T>}
 */
export const suspend = (value, f) => new Suspend(value, f)

