import { Pure, Suspend } from './free.js'
import { AlgebraWithBase, Category, Comonad, Foldable, Monoid } from './index.js'

/** @import { InferredMorphism, Morphism } from './types.js' */

const Nil = Symbol('Nil')

/**
 * @template T
 * @extends {Pure<T>}
 */
class ListPure extends AlgebraWithBase(Pure)(Category, Foldable, Monoid) {
  static empty() {
    return List.empty()
  }

  static id() {
    return List.empty()
  }

  head() {
    return this._value
  }

  tail() {
    return Empty
  }

  /**
   * @template U
   * @this {ListPure<T>}
   * @param {Morphism<T, ListPure<U>>} f 
   * @returns {ListPure<U>}
   */
  chain(f) {
    return f(this.head())
  }

  /**
   * @template U
   * @this {ListPure<T>}
   * @param {Morphism<T, U>} f 
   * @returns {ListPure<U>}
   */
  map(f) {
    return new ListPure(f(this._value))
  }


  /**
   * @template U
   * @param {(acc: U, value: T) => U} f
   * @param {U} init
   * @returns {U}
   */
  reduce(f, init) {
    return f(init, this._value)
  }

  count() {
    return this.isEmpty() ? 0 : 1
  }

  /**
   * @this {List<any>}
   * @returns {this is Empty}
   */
  isEmpty() {
    return this === Empty
  }
}

/**
 * @template T
 * @extends {Suspend<T>}
 */
class ListSuspend extends AlgebraWithBase(Suspend)(Foldable, Monoid, Comonad) {
  /**
   * @param {T} head
   * @param {() => List<T>} tail
   */
  constructor(head, tail) {
    super(
      head,
      /** @type {InferredMorphism<T>} */(tail)
    )
  }


  static empty() {
    return List.empty()
  }

  /**
   * @template T
   * @this {ListSuspend<T>}
   * @param {T} value 
   * @returns {ListSuspend<T>}
   */
  cons(value) {
    return new ListSuspend(value, () => this)
  }

  head() {
    return this._value
  }

  tail() {
    return this.step()
  }

  /**
   * @template U
   * @param {(acc: U, value: T) => U} f
   * @param {U} init
   * @returns {U}
   */
  reduce(f, init) {
    const acc = f(init, this._value)
    return this.tail().reduce(f, acc)
  }

  extract() {
    return this.reduce(reduceArray, [])
  }

  count() {
    return this.tail().count() + 1
  }

  /** @returns {this is Empty} */
  isEmpty() {
    return false
  }
}

/**
 * @template T
 * @type {ListPure<T> | ListSuspend<T>} List
 */
export class List {
  /**
   * @template  T
   * @param {T} value
   * @returns {List<T>}
   */
  static of(value) {
    return new ListSuspend(value, List.empty)
  }

  /**
   * @template T
   * @param {Iterable<T>} iterable
   * @returns {List<T>}
   */
  static from(iterable) {
    const iterator = Iterator.from(iterable)
    return List.fromIterator(iterator)
  }

  /**
   * @template T
   * @param {Iterator<T>} iterator
   * @returns {List<T>}
   */
  static fromIterator(iterator) {
    const next = iterator.next()

    if (next.done) {
      return List.empty()
    } else {
      return new ListSuspend(next.value, () => List.fromIterator(iterator))
    }
  }

  static empty() {
    return Empty
  }
}

const Empty = new ListPure(Nil)

/**
 * @template T
 * @param {T[]} acc
 * @param {T} value
 * @returns {T[]}
 */
const reduceArray = (acc, value) => acc.concat(value)

